/*

Implement a "classic" list in JavaScript
=====================================
(Grade - moderate)

JavaScript has objects and arrays, but no lists.

A list is 
	- the empty list []
	- or the concatenation of an item (head) and a list (tail) [ a . L ]
	- has methods
		- head, which returns the first element
		- tail, which returns the tail list 
		- length, which returns its length
		- cons, which takes a single argument and returns a new list whose head is the argument and tail is "this"
		- toString, which returns a string representation
		
Supplementary (1): implement "class" methods on List to build a list
	- from an array
	- from a variable number of passed arguments

Supplementary (2): implement the following list methods
	- append, which returns a new list from the elements of this followed by the elements of the second
	- toArray, returns an array representaiton of the list's elements
	- map, which takes a function parameter and calls that parameter on each element of the list, building a list from the return
	values. Function takes one parameter, the list item.
	- fold, which takes a function and an initial value for an accumulator, and
		calls the function  passing the accumulator and first element
		cals the function on each subsequent element, passing 
			the value returned from the previous call as the new accumulator 
			the list element itself
	- forEach, which takes a function parameter and calls that parameter on each element of the list, with arguments
		item: the list item
		index: the number of the item (0-based)
		list: the entire list
*/

// First a utility function to convert arguments (and other things) into an array
var $A =  function(iterable) {
  if (!iterable) return [];
  if (iterable.toArray) {
    return iterable.toArray();
  } else {
    var results = [];
    for (var i = 0, length = iterable.length; i < length; i++)
      results.push(iterable[i]);
    return results;
  }
}

// We can define a single object as the empty list
// Holds all the terminating functions for the recursive methods!
EmptyList = {
	head: 	function() 	{ return undefined; },
	tail:   	function() 	{ return undefined; },
	cons:		function(item)	{ return new List(item); },
	length:	function()		{ return 0; },
	append:	function(list)	{ return list; },
	toString: 	function()		{ return "[]"; },
	toArray:	function()		{ return []; },
	forEach:	function()		{},
	_fe:		function() 	{},
	map:		function()		{ return this; },
	fold:		function(f, a) 	{ return a; },
}

// And a regular constructor function for other lists
// Always constructed from an (optional) head and an (optional) tail
// No arguments - return empty list
// One argument - a single-element list
function List(head, tail) {
	if (!head) {
		return EmptyList;
	}
	this._head = head;
	this._tail = tail || EmptyList;
}
List.prototype = {
	head:		function() 	{ return this._head; },
	tail:		function() 	{ return this._tail; },
	cons:		function(item)	{ return new List(item, this); },
	length:	function()		{ return 1 + this._tail.length(); },
	append:	function(list)	{ return this._tail.append(list).cons(this._head); },
	toString: 	function()		{ return "[" + this._head + "." + this._tail + "]"; },
	toArray:	function()		{
		return [ this._head ].concat(this._tail.toArray());		
	},
	forEach:	function(f)	{
		this._fe(f,0,this);
	},
	_fe:		function(f,index,original) {
		f(this._head,index,original);
		this._tail._fe(f,index+1,original); 
	},
	map:		function(f)	{
		return new List( f(this._head), this._tail.map(f));
	},
	fold:		function(f, accum) {
		accum = f(accum,this._head);
		return this._tail.fold(f,accum);
	},
}

// A couple of "class methods" for list
List.fromArray = function(a)	{
	if (a.length == 0) {
		return EmptyList;
	}
	var h = a.shift();
	return new List(h,List.fromArray(a));
}
List.create = function() {
	return List.fromArray($A(arguments));
}



// Some playing around....

var empty = new List();
print(empty);

var one = empty.cons(1); 
print(empty);
print(one);

var several = List.create("one",2,"three", "iv", { five: 5 }, [ 1,2,3,4,5,6 ], "end");
print(several + " length " + several.length());

var three = List.create("x","yyyy","zzzzzzzz");
var appended = several.append(three);
print(several);
print(appended);

appended.forEach(function(item,count,list) { print(item + ", " + (count+1) + " of " + list.length()); });

var nums = List.create(1,2,3,4,5);
// list of squares
print(nums.map(function(x) { return x * x; }));
// sum of squares
print(nums.fold(function(accum, item) { return accum + item*item; },0 ));

